# Copyright (C) 2019 Catalyst Cloud Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


class InvalidConfigClass(Exception):
    pass


class ConfigGroupAlreadyRegistered(Exception):
    pass


class InvalidConfigFileFormat(Exception):
    def __init__(self, config_format=None):
        self.msg = "Cannot use unsupported config formats."
        if config_format:
            self.msg = f"'{config_format}' is not a supported config format."

    def __str__(self):
        return str(self.msg)


class BaseConfigLoaderException(Exception):
    """ """

    def __init__(self, errors=None):
        self.errors = errors

    def __str__(self):
        return str(self.errors)


class InvalidConfValue(BaseConfigLoaderException):
    pass


class InvalidConfGroup(BaseConfigLoaderException):
    pass


class InvalidConf(BaseConfigLoaderException):
    pass


class NoSuchConfig(KeyError):
    pass


class InvalidTestOperation(Exception):
    pass
